"""
This module handles global settings of the package.
"""

NUM_BYTES = 32

# Optimizations
CLASSICAL = True
POINT_AND_PERMUTE = False
GRR3 = False
FREE_XOR = False
FLEXOR = False
HALF_GATES = False
